/****************************************************************
;****************************************************************
;
;  Source code for P-45AVR-V2 & MT-10T7(8) receiver (version 0.2)
;
;    http://p-45.narod.ru/
;
;    e-mail:p-45@mail.ru
;
;***************************************************************
;
;    compile by avr-gcc with -Os options
;
;***************************************************************
;
;    distribute by GPL licence
;
;****************************************************************
;****************************************************************
*/

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>


#define F_CPU 8000000UL

//#define RC5_CONTROL  // if F_CPU == 10MHz

#include <avr/delay.h>
#include <avr/sleep.h>

#include <stdio.h>

#define SetBit(port,bit) port |=  _BV(bit)
#define ClrBit(port,bit) port &= ~_BV(bit)


#define Port4Out(port,bit) port |=  _BV(bit)
#define Port4Inp(port,bit) port &= ~_BV(bit)

#define HiBYTE(w) (uint8_t)((w >> 8) & 0x00ff)
#define LoBYTE(w) (uint8_t)((w >> 0) & 0x00ff)

#define Nop() __asm__ __volatile__ ("nop" ::)

#define CONST__GAINoff 0B10001000
#define CONST__GAINon  0B10001001 // Weak Signal Booster ON
#define CONST__LoBAND  0B00000001
#define CONST__MiBAND  0B00000010
#define CONST__HiBAND  0B00001100

#define RCVPCH1 37300
#define RCVPCH2 10700

#define RCV_MAX 855000
#define RCV_MIN  40000


#define ADR_EE_RC5    0x01A0  // 18* 2 Bytes
#define ADR_EE_vMax   0x01F0  // 1 * 1 Bytes
#define ADR_EE_vMin   0x01F1  // 1 * 1 Bytes
#define ADR_EE_vWrk   0x01F2  // 1 * 1 Bytes
#define ADR_EE_Lo     0x01F8  // 1 * 4 Bytes
#define ADR_EE_Hi     0x01FC  // 1 * 4 Bytes

#define VolumeMAX    80
#define VolumeMIN    10

#define I2C_Port_ini   DDRC
#define I2C_Port      PORTC
#define I2C_BitClc      PC5
#define I2C_BitDat      PC4


#define BUS_Port_ini   DDRB
#define BUS_Port      PORTB
#define BUS_BitClc      PB5
#define BUS_BitDat      PB3

#define Ext_Port_ini   DDRD
#define Ext_Port      PORTD
#define Ext_Load        PD5

#define KBD_Port_ini1   DDRB
#define KBD_Port_ini2   DDRC
#define KBD_Port_Inp1   PINB
#define KBD_Port_Inp2   PINC
#define KBD_Port1      PORTB
#define KBD_Port2      PORTC
#define KBD_BitInp1     PB0
#define KBD_BitInp2     PC0

#define KBDcs_Port_ini   DDRD
#define KBDcs_Port      PORTD
#define KBDcs_Load        PD7

#define LC72131_Port_ini   DDRD
#define LC72131_Port      PORTD
#define LC72131_Load        PD6

#define KBD_0    0
#define KBD_1    1
#define KBD_2    2
#define KBD_3    3
#define KBD_4    4
#define KBD_5    5
#define KBD_6    6
#define KBD_7    7
#define KBD_8    8
#define KBD_9    9
#define KBD_F    10
#define KBD_Clr  11
#define KBD_Pls  12
#define KBD_Mns  13
#define KBD_Sql  14
#define KBD_Vol  15

#define IND_Pls   13
#define IND_Mns   12

uint8_t Ind2KBD[16]={KBD_F,  KBD_7,  KBD_4,  KBD_1,
		     KBD_Clr,KBD_9,  KBD_6,  KBD_3,
		     KBD_0,  KBD_8,  KBD_5,  KBD_2,
		     KBD_Mns,KBD_Pls,KBD_Vol,KBD_Sql};

#define WFM  0
#define NFM  1
#define  AM  2

enum { UP, DOWN };

enum { RADIO,
       STEP,
       GAIN,
       UAPY,
       FINPUT,
       WR_H_L,
       WR_MEM,
       RC_MEM,
       SCAN_F,
       SCAN_M,
       SET_SQL,
       SET_VOL
     };

#define  NUM_MEM 100

#define  LcdBufLEN 10
#define  InpBufLEN  7

#define  RxInpBufLEN 16

#define  PROMPT "\n>"
#define  ERROR  "\nErr\n>"

volatile uint8_t RcInpBuf[RxInpBufLEN];
volatile uint8_t RxInpBuf[RxInpBufLEN];
volatile uint8_t RxInd = 0;
volatile uint8_t TxInd = 0;
volatile uint8_t RcInd = 0;

volatile uint8_t EchoYes = 1;

volatile uint8_t InpBuf[InpBufLEN];
	 uint8_t cccBuf[8];

volatile uint8_t InpInd;
volatile int8_t  IndMEM = 0;

	 uint8_t Step[8]={1,5,10,15,20,25,50,100};

volatile uint32_t FrcvWork;
volatile uint8_t  FrcvStep =  5;
volatile uint8_t  FindStep =  1;

volatile uint8_t  direction = UP;
volatile uint8_t  Regim;

volatile uint8_t  Modulation;

volatile uint8_t  FlagMute;
volatile uint8_t  FlagSQL = 1;
volatile uint16_t MuteCnt = 0;

volatile uint8_t Pwm1;
volatile uint8_t Pwm2;

volatile uint8_t Vol_Max;
volatile uint8_t Vol_Min;


#define  AUTO_CLR     7  // sec

volatile uint8_t AutoClr;
volatile uint8_t TimeClr;
volatile uint8_t TimeTmr;


#define  MUTE     50


volatile uint32_t FrcvMax;
volatile uint32_t FrcvMin;

	 uint8_t  FlagScanF = 0;
	 uint8_t  FlagScanM = 0;
	 uint8_t  FlagBegin = 0;

	  int8_t  ScanMBank = 0;
	 uint8_t  FlagBank;


volatile uint16_t FrcvPCH2;
volatile uint8_t  flagBo;

volatile uint8_t  Smeter;

volatile uint8_t  TunerGain  = 0;


static  void KeyNop  (void);
	void KeyMOD  (void);
	void KeyPlus (void);
	void KeyMinus(void);
	void KeyRADIO(void);
	void KeySTEP (void);
	void KeyGAIN (void);
	void KeyUapy (void);
	void KeyFinp (void);
	void KeySetHi(void);
	void KeySetLo(void);
	void KeyWrMem(void);
	void KeyRcMem(void);
	void KeyScanF(void);
	void KeyScanM(void);

	void KeySQL(void);
	void KeyVOL(void);

	void Key_0   (void);
	void Key_1   (void);
	void Key_2   (void);
	void Key_3   (void);
	void Key_4   (void);
	void Key_5   (void);
	void Key_6   (void);
	void Key_7   (void);
	void Key_8   (void);
	void Key_9   (void);

void (* SwKey[32])(void) =
     {
	KeyMOD,
	KeySTEP,
	KeyGAIN,
	KeyUapy,
	KeySetHi,
	KeyWrMem,
	KeySetLo,
	KeyScanF,
	KeyRcMem,
	KeyScanM,
	KeyFinp,
	KeyRADIO,
	KeyPlus,
	KeyMinus,
	KeySQL,
	KeyVOL,

	Key_0,
	Key_1,
	Key_2,
	Key_3,
	Key_4,
	Key_5,
	Key_6,
	Key_7,
	Key_8,
	Key_9,
	KeyFinp,
	KeyRADIO,
	KeyNop,
	KeyNop,
	KeyNop,
	KeyNop
     };


#define MT10T_Comma 0x10

#define MT10T_P     0b10101011
#define MT10T_S     0b11001101
#define MT10T_E     0b10001111
#define MT10T_B     0b11000111
#define MT10T_O     0b11101110
#define MT10T_H     0b11100011
#define MT10T_L     0b10000110
#define MT10T_U     0b11100110
#define MT10T_G     0b11001110
#define MT10T_F     0b10001011
#define MT10T_A     0b11101011
#define MT10T_N     0b01000011
#define MT10T_T     0b10000111
#define MT10T_I     0b01100000
#define MT10T_Y     0b11100101

#define MT10T_c     0b00000111
#define MT10T_q     0b11101001

#define MT10T__     0b00000100
#define MT10T_Poisk 0b11101010
#define MT10T_Minus 0b00000001

#define MT10T_WFM  0b01000101
#define MT10T_NFM  0b00000111
#define MT10T_AM   0b01000111


#define Lcd_Comma MT10T_Comma
#define Lcd_P     MT10T_P
#define Lcd_S     MT10T_S
#define Lcd_T     MT10T_T
#define Lcd_E     MT10T_E
#define Lcd_N     MT10T_N
#define Lcd_G     MT10T_G
#define Lcd_A     MT10T_A
#define Lcd_B     MT10T_B
#define Lcd_I     MT10T_I
#define Lcd_F     MT10T_F
#define Lcd_O     MT10T_O
#define Lcd_U     MT10T_U
#define Lcd_H     MT10T_H
#define Lcd_L     MT10T_L
#define Lcd_Y     MT10T_Y
#define Lcd_c     MT10T_c
#define Lcd_q     MT10T_q
#define Lcd__     MT10T__
#define Lcd_Poisk MT10T_Poisk
#define Lcd_Minus MT10T_Minus

#define Lcd_WFM   MT10T_WFM
#define Lcd_NFM   MT10T_NFM
#define Lcd_AM    MT10T_AM

#define Lcd4i     5   // 4
#define Lcd4f     3   // 2

	void MT10TxInit(void);
	void MT10TxLcd (void);

	void BusOutH2L(uint8_t Bus);
	void BusOutL2H(uint8_t Bus);

	void    cMEM(void);
	void    wMEM(void);
	uint8_t rMEM(void);

	void ScanerF(void);
	void ScanerM(void);
	void SetMute(void);

	void RsControl  (void);
	void UapyControl(void);

	void VolumeUp  (void);
	void VolumeDown(void);

	void uart_putstr(char *p);
	int  uart_putchar(char c);

	void wBufEEPROM(uint16_t addr, uint8_t *buff, uint8_t size);
	void rBufEEPROM(uint16_t addr, uint8_t *buff, uint8_t size);

	void SQLup  (void);
	void SQLdown(void);

	void PDUsqlUp  (void);
	void PDUsqlDown(void);

	void PDUvolUp  (void);
	void PDUvolDown(void);

	void UapyUp  (void);
	void UapyDown(void);

	  uint8_t * ul2a(uint32_t c,uint8_t len);
	  uint32_t  a2l(uint8_t *p);

void (* RcCom[4])(void) =
     {
	PDUsqlUp,     // 14  'W' code for control by RS-232
	PDUsqlDown,   // 15
	PDUvolUp,     // 16
	PDUvolDown    // 17
     };

/////////////////////////////////////////////////
void SetFmax(void) { FrcvMax = RCV_MAX; return; }
void SetFmin(void) { FrcvMin = RCV_MIN; return; }

/////////////////
void KeyNop(void)
{
  return;
}

//////////////////////////
//////////////////////////
volatile uint8_t  BeepCnt;
volatile uint8_t  BeepTmr;

#define BEEP_Port_ini   DDRB
#define BEEP_Port      PORTB
#define BEEP_Bit         PD4

#define BEEP_T      25  // ms      // fix ME
#define BEEP_F     900  // Hz

#define BEEP_TMR  (7812/BEEP_F)
#define BEEP_LEN  (255-BEEP_T*15/BEEP_TMR)


//////////////////////////////////////////
void Beep(void) { BeepCnt = BEEP_LEN; return; }

/////////////////
void Beeper(void)
{
   if(++BeepTmr > BEEP_TMR)
   {
	BeepTmr = 0;

     if(BeepCnt)
     {
	if(++BeepCnt & 0x01)
	{
	   SetBit(BEEP_Port,BEEP_Bit);
	}
	else
	   ClrBit(BEEP_Port,BEEP_Bit);
     }
   }
   return;
}

///////////////////
////// U-apy //////
///////////////////

#define UapyNN  40
#define UapyHi  235
#define UapyLo  210

volatile uint8_t  Uapy = 0xff;
volatile uint8_t  UapyFlg = 1;
	 uint8_t  UapySet = 0xff;
	 uint8_t  UapyCnt;
	 uint16_t SmeterUapy = 0;

//////////////////
void UapyOut(void)
{
     OCR1B = Uapy;
}

////////////////////////////
// Port Extention 74HC595 //

#define EP0 0
#define EP1 1
#define EP2 2
#define EP3 3
#define EP4 4
#define EP5 5
#define EP6 6
#define EP7 7


///////////////////////////////
void BusPause(void) { return; }

///////////////////
void BusStrob(void)
{
     BusPause();
	  SetBit(BUS_Port,BUS_BitClc);
     BusPause();
	  ClrBit(BUS_Port,BUS_BitClc);
}

///////////////////////////
void BusOutH2L(uint8_t Bus)
{
  uint8_t i;

    for(i=0; i < 8; i++)
    {
      if(Bus & 0x80)
	 SetBit(BUS_Port,BUS_BitDat);
      else
	 ClrBit(BUS_Port,BUS_BitDat);

      BusStrob();

      Bus <<= 1;
    }
    return;
}

///////////////////////////
void BusOutL2H(uint8_t Bus)
{
  uint8_t i;

    for(i=0; i < 8; i++)
    {
      if(Bus & 0x01)
	 SetBit(BUS_Port,BUS_BitDat);
      else
	 ClrBit(BUS_Port,BUS_BitDat);

      BusStrob();

      Bus >>= 1;
    }
    return;
}

/////////////////////
void KBD_cs(void)
{
     BusPause();
	  SetBit(KBDcs_Port,KBDcs_Load);
     BusPause();
	  ClrBit(KBDcs_Port,KBDcs_Load);

     return;
}


volatile uint8_t  ExtPort = 0;

/////////////////////
void ExtPortSet(void)
{
     BusOutH2L(ExtPort);

     BusPause();
	  SetBit(Ext_Port,Ext_Load);
     BusPause();
	  ClrBit(Ext_Port,Ext_Load);
     return;
}

///////////////////////////////////////////////////////////////////
void SetExtBit(uint8_t bit) { ExtPort |=  _BV(bit); ExtPortSet(); }
void ClrExtBit(uint8_t bit) { ExtPort &= ~_BV(bit); ExtPortSet(); }

///////////////////////////
uint8_t LcdDigit(uint8_t c)
{
  static char MT10TxDigit[]={
	0b11101110,     // 0
	0b01100000,     // 1
	0b00101111,     // 2
	0b01101101,     // 3
	0b11100001,     // 4
	0b11001101,     // 5
	0b11001111,     // 6
	0b01101000,     // 7
	0b11101111,     // 8
	0b11101101,     // 9
	0b11101011,     // A
	0b11000111,     // B
	0b10001110,     // C
	0b01100111,     // D
	0b10001111,     // E
	0b10001011};    // F

	return(MT10TxDigit[(c & 0x0F)]);
}

volatile uint8_t  LcdWork = 1;
volatile uint8_t  LcdBuf [LcdBufLEN];

volatile uint8_t  LcdYes;
volatile uint8_t  LcdCnt;
volatile uint8_t  LcdFlg;
volatile uint16_t LcdBlink;

/////////////////
void LcdTmr(void)
{
   if(++LcdCnt > 50)
   {
	LcdCnt = 0;
	LcdYes++;

     if(LcdBlink)
     {
	  LcdFlg = (LcdFlg)? 0:1;
     }
     else LcdFlg = 0;
   }
   return;
}


/////////////////////////////
//          KBD            //
/////////////////////////////

volatile uint8_t  KbdPos = 0;
volatile uint8_t  KbdKbd = 0;
volatile uint8_t  KbdYes = 0;
volatile uint8_t  KbdInd;
volatile uint8_t  KbdCnt;
volatile uint8_t  KbdOld;
volatile uint8_t  KbdFst;

volatile uint8_t  TmrInd;

/////////////////
void KbdWrk(void)
{
    if(KbdCnt < 128)
       KbdCnt++;

    if( bit_is_clear(KBD_Port_Inp1,KBD_BitInp1) ||
	bit_is_clear(KBD_Port_Inp2,KBD_BitInp2) )
    {
       BusOutH2L(~_BV(KbdPos)); KBD_cs();

       Nop();
       Nop();
       Nop();
       KbdInd = 255;

       if(bit_is_clear(KBD_Port_Inp1,KBD_BitInp1)) KbdInd = KbdPos;
       if(bit_is_clear(KBD_Port_Inp2,KBD_BitInp2)) KbdInd = KbdPos+8;

       if(KbdInd < 255)
       {
	 if(KbdYes && KbdOld == KbdInd)
	 {
	    if( KbdInd == IND_Pls ||
		KbdInd == IND_Mns )
	    {
	      if(KbdCnt > 50)
	      {
		 KbdKbd = Ind2KBD[KbdInd]+1;
		 KbdFst = 1;
	      }
	    }
	 }
	 else
	 {
	    KbdYes = 1;
	    KbdCnt = 0;
	    KbdFst = 0;
	    KbdOld = KbdInd;
	    KbdKbd = Ind2KBD[KbdInd]+1;
	 }
       }
       BusOutH2L(0x00); KBD_cs();
    }
    else
    {
       KbdYes = 0;
    }

    if(++KbdPos >= 8) KbdPos = 0;
}

////////////////////
uint8_t GetKbd(void)
{
  uint8_t tmp;

      //cli();
      tmp = KbdKbd;

   if(tmp)
   {
      KbdKbd = 0;
   }
   //sei();
   return(tmp);
}


#ifdef  RC5_CONTROL
///////////////////////////
//      RC-5 decoder     //
///////////////////////////

#define RC5_Port_ini   DDRD
#define RC5_Port_Inp   PIND
#define RC5_Port      PORTD
#define RC5_BitInp      PD2

volatile uint8_t RC5flg;
volatile uint8_t RC5err;
volatile uint8_t RC5num;
volatile uint8_t RC5cnt;
volatile uint8_t RC5wrk;
volatile uint8_t RC5prg;

volatile uint16_t RC5inp;
volatile uint16_t RC5old;
volatile uint16_t RC5rcv;
volatile uint16_t RC5tmp;

volatile  int8_t RC5cnt0a;
volatile  int8_t RC5cnt1a;
volatile  int8_t RC5cnt0b;
volatile  int8_t RC5cnt1b;

////////////////////////
//   for 10.0 MHZ
//
#define  RC5_CONST_E  6
#define  RC5_CONST_1  33
#define  RC5_CONST_2  67

/////////////////////////
//  Interrupt by INT0  //
/////////////////////////
SIGNAL (SIG_INTERRUPT0)
{
   if(!RC5flg)
   {
       RC5flg++;
       RC5cnt = 0;
       RC5rcv = 0;
       RC5num = 0;
       RC5err = 0;
   }
   return;
}

////////////////////
void RC5AskBit(void)
{
  uint8_t a;
  uint8_t b;

     if(++RC5num < 14)
     {
     if(abs(RC5cnt1a-RC5cnt0a) < RC5_CONST_E) RC5err++;
     if(abs(RC5cnt1b-RC5cnt0b) < RC5_CONST_E) RC5err++;

       a = (RC5cnt1a > RC5cnt0a)? 1:0;
       b = (RC5cnt1b > RC5cnt0b)? 1:0;

       if(a == b) RC5err++;

	RC5rcv <<= 1;
	RC5rcv  |= (a > b)? 1:0;
     }
     else
     {
	if(RC5num > 14)
	{
	   if(!RC5err && RC5rcv & 0x1000)
	   {
	       RC5inp =  RC5rcv;
	   }
	   RC5flg = 0;
	}
     }
     RC5cnt1a =
     RC5cnt0a =
     RC5cnt1b =
     RC5cnt0b = 0;
     return;
}

/*
///////////////////////////////
//
//  For 4.0 MHZ
//
uint8_t RC5corrector(uint8_t i)
{
  static char corrector[12]={26,26,25,26,
			     26,26,25,26,
			     26,26,25,26};

  return((i > 11)? 26:corrector[i]);
}
*/

//////////////////////
void RC5receiver(void)
{
     if(RC5flg)
     {
	if(RC5cnt > RC5_CONST_1)
	{
	   if(bit_is_clear(RC5_Port_Inp,RC5_BitInp))
	   {
		RC5cnt0b++;
	   }
	   else
		RC5cnt1b++;
	}
	else
	{
	   if(bit_is_clear(RC5_Port_Inp,RC5_BitInp))
	   {
		RC5cnt0a++;
	   }
	   else
		RC5cnt1a++;
	}
	if(RC5cnt++ > RC5_CONST_2)
	{
	   RC5cnt = 0;
	   RC5AskBit();
	}
     }
     return;
}


////////////////////////
void RC5write(uint8_t c)
{
     if(c < 18)
     {
       if(RC5prg > 3)
       {
	  RC5prg = 0;

	  wBufEEPROM ((uint16_t)(c<<2) + ADR_EE_RC5, (uint8_t *)&RC5old, 2);
	  uart_putstr("\nOK");
       }
     }
     return;
}

////////////////////////////
uint16_t RC5_Kbd(void)
{
  uint8_t i;

     for(i = 0; i < 18; i++)
     {
       rBufEEPROM((uint16_t)(i<<2) + ADR_EE_RC5, (uint8_t *)&RC5tmp, 2);

       if( RC5tmp == (RC5inp & 0x07ff) )
       {
	 if(RC5old != RC5inp || i >= KBD_Pls)
	 {
	    if(i < 14)
	    {
	       Beep();
	       RC5old = RC5inp;  return i+1;
	    }
	    else
	    {
	      (* RcCom[i-14])(); return 0;
	    }
	 }
       }
     }
     return 0;
}
#endif // RC5_CONTROL

/////////////////////////////
/////////////////////////////
     volatile  int8_t TMRxxx;

void TmrYYY(void)
{
  if(TMRxxx)
  {
     TMRxxx = 0;

     KbdWrk();
     MT10TxLcd();
  }
  return;
}

/////////////////////////
// Interrupt by Timer0 //
/////////////////////////
//
// Start by 64.0 mks - for  4.0 MHZ
//          25.6 mks - for 10.0 MHZ *
//
SIGNAL (SIG_OVERFLOW0)
{
#ifdef RC5_CONTROL
  RC5receiver();
#endif

  Beeper();

  if(TmrInd <  254)
     TmrInd++;

  if(TmrInd == 200)
  {
     TmrInd = 0;
     TMRxxx++;

     LcdTmr();

     if(MuteCnt < MUTE)
	MuteCnt++;

     if(++TimeTmr == 0)   // 1048 ms for 4.0 MHZ ;;; 500.2ms for 10.0 MHZ
     {
       if(TimeClr)
       {
	  if(--TimeClr ==  0) AutoClr++;
       }
     }
  }
}

////////////////////////////
// Interrupt by Rx RS-232 //
////////////////////////////
SIGNAL (SIG_UART_RECV)
{
  uint8_t tmp;

  if(RxInd < RxInpBufLEN)
  {
     RxInpBuf[RxInd++] = UDR;
  }
  else tmp = UDR;
}

//--------------//
void ioinit (void)
{
    // Init PORTs

    Port4Out(BUS_Port_ini,BUS_BitClc); // Bus clc -> Out
    Port4Out(BUS_Port_ini,BUS_BitDat); // Bus dat -> Out

    Port4Out(Ext_Port_ini,Ext_Load);   // Port Extention 74HC595 load -> Out

    Port4Out(LC72131_Port_ini,LC72131_Load); // Port PLL_CS -> Out
    Port4Out(KBDcs_Port_ini,KBDcs_Load);     // Port KBD_CS -> Out

    Port4Inp(KBD_Port_ini1,KBD_BitInp1); // KBD inp1 -> Input
    Port4Inp(KBD_Port_ini2,KBD_BitInp2); // KBD inp2 -> Input
    SetBit  (KBD_Port1,KBD_BitInp1);     //     Input (+Ucc --> "1")
    SetBit  (KBD_Port2,KBD_BitInp2);     //     Input (+Ucc --> "1")

    Port4Out(I2C_Port_ini,I2C_BitClc); // I2C clc -> Out
    Port4Out(I2C_Port_ini,I2C_BitDat); // I2C dat -> Out

    SetBit(I2C_Port,I2C_BitClc);  // Init I2C
    SetBit(I2C_Port,I2C_BitDat);

    // BEEPER

    Port4Out(BEEP_Port_ini,BEEP_Bit);  // Beeper -> Out

    // Init ADC

    Port4Inp(DDRC,PC0);
    Port4Inp(DDRC,PC1);
    Port4Inp(DDRC,PC2);
    Port4Inp(DDRC,PC3);

    ADMUX  = 0x00;
    ADCSRA = 0x85; // 1000 0101  ADC on, interrupt disable, MCU clock / 32

    // Init Timer0

    TCCR0  = 0x01;       // Timer0 running on MCU clock / 1
    SetBit(TIMSK,TOIE0); // enable interrupt from Timer0


    // Init Timer1

    Port4Out(DDRB,PB1); // output PWM1 (PB1/OC1A)
    Port4Out(DDRB,PB2); // output PWM2 (PB2/OC1B)

    TCCR1A = 0B10100001; // Timer1 is duble 8-bit PWM
    TCCR1B = 0B00000001; // Timer1 running on MCU clock / 1

    OCR1A = Pwm1 = 255;  // set PWM1 value
    OCR1B        = 255;  // set PWM2 value

#ifdef RC5_CONTROL
    // Init INT0

    Port4Inp(RC5_Port_ini,RC5_BitInp); // RC5 inp -> Input
    SetBit  (RC5_Port,    RC5_BitInp); // RC5 Input (+Ucc --> "1")
		       //         _
    MCUCR = 0x03;      // int0  _|   (rising edge)

    SetBit(GICR,INT0); // enable interrupt from INT 0
#endif

    // ### RS-232 ###

    Port4Inp(DDRD,PD0); // RS-232 Rx (Inp)
    Port4Out(DDRD,PD1); // RS-232 Tx (Out)

    UBRRL = (F_CPU / (16 * 9600UL)) - 1;        // Bd 9600
    UCSRB = _BV(TXEN) | _BV(RXEN) | _BV(RXCIE); // Tx,Rx,Rx interrupt enable

    // enable interrupts
       sei ();
       return;
}

///////////////////////////////////////
void Pause200(void) { _delay_us(200); }
void Pause10 (void) { _delay_us( 10); }

/////////////////////
void sleep(int16_t s)
{
  for(s*=5; s > 0; s--)
  {
  TmrYYY();
      Pause200();
  }
}


////////////////////////
int uart_putchar(char c)
{
  if (c == '\n')
     uart_putchar('\r');

  while(bit_is_clear(UCSRA, UDRE))
  {
    //TmrYYY();
     Nop();
  }
  UDR  = c;
  return 0;
}

////////////////////////
void uart_putstr(char *p)
{
   while(*p)
   {
     uart_putchar(*p++);
   }
   return;
}

////////////////////////

void PutHEX(uint8_t c)
{
  static char hex[16]= {'0','1','2','3','4','5','6','7',
			'8','9','A','B','C','D','E','F'};

 uart_putchar('[');

   uart_putchar(hex[c >>   4]);
   uart_putchar(hex[c & 0x0f]);

 uart_putchar(']');
   return;
}

//////////////////
void Rcv2Lcd(void)
{
  char *p;

  uint8_t i;
  uint8_t flag = 0;

    p = ul2a(FrcvWork,6);

    for(i=0; i < 6; i++)
    {
      cccBuf[i]=*p;

      if(*p != '0') flag++;

	       LcdBuf[i+Lcd4f] &= Lcd_Comma;
      if(flag) LcdBuf[i+Lcd4f] |= LcdDigit(*p-'0');

      p++;
    }
    return;
}


////////////////////////////////////
void Int2Lcd(uint16_t c, uint8_t sw)
{
  char *p;

  uint8_t i;
  uint8_t flag = 0;

    p = ul2a(c,4);

    for(i=0; i < 4; i++)
    {
      if(*p != '0' || i == 3) flag++;

	       LcdBuf[i+sw] &= Lcd_Comma;
      if(flag) LcdBuf[i+sw] |= LcdDigit(*p-'0');

      p++;
    }
    return;
}


//////////////////
void Ind2Lcd(void)
{
  uint8_t *p;
  uint8_t  i;

    p = ul2a(IndMEM,2);

    for(i=0; i < 2; i++)
    {
       LcdBuf[i] &= Lcd_Comma;
       LcdBuf[i] |= LcdDigit(*p-'0');
       p++;
    }
    return;
}

////////////////////////////////
//  Analog-Digital Converter  //
////////////////////////////////

#define CH_VOLUME 0x42
#define CH_POROG  0x41
#define CH_SMETR  0x43
#define CH_GND    0x4f

//--------------//
void ADCwrk(void)
{
  Pause200();
  SetBit(ADCSRA,ADSC); // start ADC

  while(bit_is_set(ADCSRA,ADSC))
  {
    TmrYYY();
     //Nop();
  }
}

//--------------//
uint16_t GetADC (uint8_t chanel)
{
  ADMUX = chanel;

  ADCwrk();
  ADCwrk();

  return(ADCW);
}

/////////////////////////////
void SmeterLcd(uint8_t level)
{
   int8_t i;
  uint8_t f = 0;

     for(i = 7; i >= 0; i--)
     {
	if(level & 0x80) f++;

	      LcdBuf[i] &= ~Lcd_Comma;
	if(f) LcdBuf[i] |=  Lcd_Comma;

	level <<= 1;
     }
     return;
}

////////////////////
void SmeterPrn(void)
{
  uint8_t *p;
  uint8_t  i;

  uint16_t level = LoBYTE(GetADC(CH_SMETR)>>2);

    SmeterLcd(level);

    uart_putchar(' ');

    p = ul2a(level,4);

    for(i=0; i < 4; i++)
    {
	uart_putchar(*p++);
    }
    return;
}


////////////
void wEEPROM(uint16_t addr, uint8_t byte)
{
     while(bit_is_set(EECR,EEWE))
     {
       Nop();
     }
     EEAR = addr;
     EEDR = byte;
     cli ();
	 SetBit(EECR,EEMWE);
	 SetBit(EECR,EEWE ); // start write to eeprom
     sei ();
     return;
}

///////////////
uint8_t rEEPROM(uint16_t addr)
{
     while(bit_is_set(EECR,EEWE))
     {
       Nop();
     }
     EEAR = addr;
     SetBit(EECR,EERE);
     return(EEDR);
}

///////////////
void wBufEEPROM(uint16_t addr, uint8_t *buff, uint8_t size)
{
     uint8_t i;

     for(i=0; i < size; i++)
     {
       wEEPROM(addr++,*buff); buff++;
     }
     return;
}

///////////////
void rBufEEPROM(uint16_t addr, uint8_t *buff, uint8_t size)
{
     uint8_t i;

     for(i=0; i < size; i++)
     {
	*buff = rEEPROM(addr++);
	 buff++;
     }
     return;
}

/////////////////////////
void SetFwork(uint32_t f)
{
     if( f <=  FrcvMax &&
	 f >=  FrcvMin )
     {
	 if(FindStep)
	    FrcvWork = (f/5)*5;
	 else
	    FrcvWork = f;
     }
     KeyRADIO();
}


#define LC72131_IN1 0x14 // 00010100
#define LC72131_IN2 0x94 // 10010100

////////////////////
void Lc72131Set(void)
{
  static uint8_t  Mod4Lc72131[3] = {0xC0,0xC2,0xC1};
	 uint8_t  dev = 1;
	 uint8_t  cnt = 0x9D;
	 uint16_t fPCH2;

     if(FindStep)
     {
       dev = 5;
       cnt = 0x95;
     }

     FrcvPCH2 = (RCVPCH1-RCVPCH2)-(FrcvWork-((FrcvWork/50)*50));
	fPCH2 = FrcvPCH2/dev;

     BusOutH2L(LC72131_IN1);

     SetBit(LC72131_Port,LC72131_Load);

     BusOutL2H(LoBYTE(fPCH2));
     BusOutL2H(HiBYTE(fPCH2));

	  ///////////////////////////////////////
	  //
	  //  -- Lc72131 IN1 --
	  //
	  ///////////////////////////////////////
	  //
	  //  SNS DVS CTE XS R0 R1 R2 R3
	  //
	  //  R0=0, R1=0, R2=1, R3=0  --> 12,5kHz
	  //  R0=0, R1=1, R2=0, R3=1  -->  5  kHz
	  //  R0=1, R1=1, R2=0, R3=1  -->  1  kHz
	  //
	  //  SNS=1,DVS=0 --> AMIN
	  //
	  //  XS=1 --> 7,2 mHz
	  //
	  //  CTE=1 --> IF counter start
	  //  CTE=0 --> IF counter reset
	  //
	  //  10010010 0x92 -> 12,5 kHz
	  //  10010101 0x95 ->  5,0 kHz
	  //  10011101 0x9D ->  1,0 kHz

     BusOutH2L(cnt);  // 0x95 -> 5kHz

     ClrBit(LC72131_Port,LC72131_Load);

	  ///////////////////////////////////////
	  //
	  //  -- Lc72131 IN2 --
	  //
	  ///////////////////////////////////////
	  //  IOC1 IOC2 IO1 IO2 BO1 BO2 BO3 BO4
	  //
	  //  BO1 BO2 BO3 BO4 = 0 -> open
	  //  BO1 BO2 BO3 BO4 = 1 -> low
	  //
	  //  IOC1 IOC2 = 0 -> input  mode IO1 IO2
	  //  IOC1 IOC2 = 1 -> output mode IO1 IO2
	  //
	  //  11000001 --> 0xC1
	  //
	  ////////////////////////////////////////
	  //  DNC DOC0 DOC1 DOC2 UL0 UL1 DZ0 DZ1
	  //
	  //  0   0    0    0    0   0   0    0
	  //
	  ////////////////////////////////////////
	  //  GT0 GT1 TBC DLC IFS TEST0 TEST1 TEST2
	  //
	  //  0   0   0   x   1   0     0     0

     BusOutH2L(LC72131_IN2);

      SetBit(LC72131_Port,LC72131_Load);

      flagBo = Mod4Lc72131[Modulation];

      if(!FlagMute) { flagBo |= 0x04; }

     BusOutH2L(flagBo);
     BusOutH2L(0x00);
     BusOutH2L(0x08);

     ClrBit(LC72131_Port,LC72131_Load);

     return;
}


/////////////////
//   I2C BUS   //
/////////////////

// Fix me -> need hardware support i2c bus

///////////////////
void I2CStart(void)
{
     SetBit(I2C_Port,I2C_BitClc);
     SetBit(I2C_Port,I2C_BitDat);

    Pause10();
     ClrBit(I2C_Port,I2C_BitDat);

    Pause10();
     ClrBit(I2C_Port,I2C_BitClc);
}

//////////////////
void I2CStop(void)
{
     ClrBit(I2C_Port,I2C_BitDat);
     SetBit(I2C_Port,I2C_BitClc);

    Pause10();
     SetBit(I2C_Port,I2C_BitDat);
}

//////////////////
void I2CSync(void)
{
    Pause10();
     SetBit(I2C_Port,I2C_BitClc);

    Pause10();
     ClrBit(I2C_Port,I2C_BitClc);
}

//////////////////////////
void I2COutByte(uint8_t c)
{
  uint8_t i;

    for(i=0; i < 8; i++)
    {

      if(c & 0x80)
	 SetBit(I2C_Port,I2C_BitDat);
      else
	 ClrBit(I2C_Port,I2C_BitDat);

	I2CSync();

	c <<= 1;
    }
    I2CSync(); // ASK
    return;
}

////////////////////
void KSH14xPLL(void)
{
  uint32_t fPCH1;
  uint8_t   Band;


     Band = CONST__LoBAND;

     if(FrcvWork > 155000L) Band = CONST__MiBAND;
     if(FrcvWork > 440000L) Band = CONST__HiBAND;


     fPCH1 = (uint16_t)((FrcvWork + RCVPCH1)/50);

     I2CStart();

     I2COutByte(0b11000000);   // Tuner Address
     I2COutByte(HiBYTE(fPCH1));
     I2COutByte(LoBYTE(fPCH1));
     I2COutByte((TunerGain)? CONST__GAINon:CONST__GAINoff);
     I2COutByte(Band);

     I2CStop();
     return;
}

//////////////////
void fWorkPrn(void)
{
  uint8_t i;

     Rcv2Lcd(); sleep(5);

     uart_putstr("\nF ");

     GetADC(CH_SMETR);

     for(i = 0; i < 6; i++)
     {
	uart_putchar(cccBuf[i]);

	if(i == 2)
	      uart_putchar('.');
     }
     SmeterPrn();
     SmeterPrn();
     SmeterPrn();
     return;
}

//////////////////
void SetFrcv(void)
{
     Lc72131Set();
     KSH14xPLL();

     fWorkPrn();
     return;
}

// --------- //
int main (void)
{
  uint8_t key;

    ioinit ();

    MT10TxInit();

    BusOutH2L(0x00); KBD_cs();

    SetFmax();
    SetFmin();

    rBufEEPROM (ADR_EE_vMax, (uint8_t *)&Vol_Max, 1);
    rBufEEPROM (ADR_EE_vMin, (uint8_t *)&Vol_Min, 1);

    rBufEEPROM (ADR_EE_vWrk, (uint8_t *)&Pwm1, 1);

    if(Vol_Max == 0xff || Vol_Min == 0xff)
    {
       Vol_Max = VolumeMAX;
       Vol_Min = VolumeMIN;
    }
    if(Pwm1 == 0xff || Pwm1 > Vol_Max || Pwm1 < Vol_Min)
    {
       Pwm1 = Vol_Max;
    }
    OCR1A = Pwm1;

    FlagMute = 1;
    Modulation = WFM;
    FrcvWork   = 106600;

    rMEM();     // IndMEM = 0

    KeyRADIO(); //sleep(1000);

#ifdef RC5_CONTROL
    RC5wrk = 1;
#endif

    KSH14xPLL();

    for (;;)
    {
      TmrYYY();

      RsControl();

      if(AutoClr)
      {
	 AutoClr = 0; KeyRADIO();
      }

      key = GetKbd();

#ifdef RC5_CONTROL

      if(RC5inp)
      {
	if(RC5wrk)
	{
	   key = RC5_Kbd();
	}
	else
	{
	   uart_putstr("RC5:");
	   PutHEX(HiBYTE(RC5inp));
	   PutHEX(LoBYTE(RC5inp));
	   uart_putstr(PROMPT);

	   RC5inp &= 0x07ff;

	   if(RC5inp == RC5old)
	   {
	      RC5prg++;
	   }
	   else
	   {
	      RC5prg = 0;
	      RC5old = RC5inp;
	   }
	}
	RC5inp=0;
      }
#endif // RC5_CONTROL

      if(key--)
      {
	if(!KbdFst)
	    Beep();

	(* SwKey[((key) & 0x0f) | ((Regim == FINPUT)? 0x10:0x00)] )();

	LcdYes++;
      }

      if(FlagScanF) { ScanerF(); LcdYes++; }
      if(FlagScanM) { ScanerM(); LcdYes++; }

      SmeterLcd(Smeter = LoBYTE(GetADC(CH_SMETR)>>2));

      SetMute();

      UapyControl();
/*
      while(!LcdYes && !TMRxxx && RxInd == TxInd)
      {
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_mode();
      }
*/
    }
}

//////////////////
void FrcvUpDown(void)
{
    switch(direction)
    {
      case UP:   if(FrcvWork <  FrcvMax)
		 {
		    FrcvWork += FrcvStep;
		 }
		 else
		    direction = DOWN;

		 break;

      case DOWN: if(FrcvWork >  FrcvMin)
		 {
		    FrcvWork -= FrcvStep;
		 }
		 else
		    direction = UP;

		 break;
    }
    return;
}


//////////////////
void KeyPlus(void)
{
   if(Regim == RADIO)
   {
     direction = UP;

     FrcvUpDown();
     SetFrcv();
   }

   if(Regim == WR_MEM || Regim == RC_MEM)
   {
      if(++IndMEM > NUM_MEM-1)
	   IndMEM = 0;
      Ind2Lcd();

      if(Regim == RC_MEM) rMEM();
   }

   if(Regim == SCAN_M && FlagBank == 0)
   {
      if(++ScanMBank > 3)
	   ScanMBank = 0;

      Int2Lcd(ScanMBank,Lcd4i);
   }

   if(Regim == SET_SQL) SQLup();
   if(Regim == SET_VOL) VolumeUp();
   if(Regim == UAPY   ) UapyUp();

   return;
}

//////////////////
void KeyMinus(void)
{
   if(Regim == RADIO)
   {
     direction = DOWN;

     FrcvUpDown();
     SetFrcv();
   }

   if(Regim == WR_MEM || Regim == RC_MEM)
   {
      if(--IndMEM < 0)
	   IndMEM = NUM_MEM-1;
      Ind2Lcd();

      if(Regim == RC_MEM) rMEM();
   }

   if(Regim == SCAN_M && FlagBank == 0)
   {
      if(--ScanMBank < 0)
	   ScanMBank = 3;

      Int2Lcd(ScanMBank,Lcd4i);
   }

   if(Regim == SET_SQL) SQLdown();
   if(Regim == SET_VOL) VolumeDown();
   if(Regim == UAPY   ) UapyDown();

   return;
}


///////////////////////////////////////////
static uint8_t On2Lcd[3] = {    0,Lcd_O,Lcd_N};
static uint8_t Of2Lcd[3] = {Lcd_O,Lcd_F,Lcd_F};

/////////////////////////
void Flg2Lcd( uint8_t sw)
{
     memcpy((uint8_t *)&LcdBuf[6],(sw)? On2Lcd:Of2Lcd,3);
     return;
}

/////////////////////////////////////////////////////////
static uint8_t Gain2Lcd[6] = {Lcd_G,Lcd_A,Lcd_I,Lcd_N,0,0};
static uint8_t Uapy2Lcd[6] = {Lcd_U,Lcd_A,Lcd_P,Lcd_Y,0,0};
static uint8_t Banc2Lcd[6] = {Lcd_B,Lcd_A,Lcd_N,Lcd_c,0,0};
static uint8_t Step2Lcd[6] = {Lcd_S,Lcd_T,Lcd_E,Lcd_P,0,0};
static uint8_t _sql2Lcd[6] = {Lcd_S,Lcd_q,Lcd_L,0,0,0};
static uint8_t _vol2Lcd[6] = {Lcd_U,Lcd_O,Lcd_L,0,0,0};


#define sGain 0
#define sUapy 1
#define sBanc 2
#define sStep 3
#define sSql  4
#define sVol  5

void Str2Lcd(uint8_t sw)
{
  uint8_t *p;

	if(sw == sGain) p = Gain2Lcd;
   else if(sw == sUapy) p = Uapy2Lcd;
   else if(sw == sBanc) p = Banc2Lcd;
   else if(sw == sStep) p = Step2Lcd;
   else if(sw == sSql ) p = _sql2Lcd;
   else                 p = _vol2Lcd;

   memcpy((uint8_t *)LcdBuf,p,6);
}

//////////////////
void KeySTEP(void)
{
     switch(Regim)
     {
       case STEP:  if(++FindStep >= 8)
			FindStep  = 0;
		   FrcvStep = Step[FindStep];
       case RADIO:
		   Str2Lcd(sStep);
		   Int2Lcd(FrcvStep,Lcd4i);

		   Regim = STEP;
     }
     return;
}

//////////////////
void Mod2Lcd(void)
{
  static uint8_t Mod4Lcd[3] = {Lcd_WFM,Lcd_NFM,Lcd_AM};

  LcdBuf[2] = Mod4Lcd[Modulation];
}

//////////////////
void KeyMOD(void)
{
   if(Regim == RADIO)
   {
     if(++Modulation > AM)
	  Modulation = WFM;

     Mod2Lcd();
     Lc72131Set();
   }
   return;
}

//////////////////
void KeyRADIO(void)
{
     if(Regim == SET_VOL)
     {
	wBufEEPROM (ADR_EE_vWrk, (uint8_t *)&Pwm1, 1);
     }

     if(Regim == SCAN_F)
     {
	FlagScanF = 0;
	SetFmax();
	SetFmin();
     }
     FlagScanM = 0;
     Regim = RADIO;

     TimeClr = 0;

     LcdBuf[1] = Lcd_P;
     LcdBuf[0] = 0;
     LcdBlink  = 0;

     Mod2Lcd();
     SetFrcv();
}


//////////////////
void KeyGAIN(void)
{
     switch(Regim)
     {
       case GAIN:  TunerGain = (TunerGain)? 0:1;
		   KSH14xPLL();

       case RADIO:
		   Str2Lcd(sGain);
		   Flg2Lcd(TunerGain);

		   Regim = GAIN;
     }
     return;
}

/////////////////
void UapyUp(void)
{
     if(UapyFlg == 2)
     {
	if(Uapy < 255)
	   Uapy++;

	UapyOut();
	Int2Lcd(Uapy,Lcd4i);
     }
     return;
}
///////////////////
void UapyDown(void)
{
     if(UapyFlg == 2)
     {
	if(Uapy)
	   Uapy--;

	UapyOut();
	Int2Lcd(Uapy,Lcd4i);
     }
     return;
}

//////////////////
void KeyUapy(void)
{
     switch(Regim)
     {
       case UAPY:
			if(UapyFlg == 1)
			{
			   UapyFlg = 2;
			   Uapy = UapySet;
			}
		   else if(UapyFlg == 2)
			{
			   UapyFlg = 0;
			   UapySet = Uapy;
			   Uapy = 0xff;
			}
		   else {
			   UapyFlg = 1;
			}
			UapyOut();

       case RADIO:
		   Str2Lcd(sUapy);

		   if(UapyFlg != 2)
		   {
		      Flg2Lcd(UapyFlg);
		   }
		   else
		   {
		      Int2Lcd(Uapy,Lcd4i);
		   }
		   Regim = UAPY;
     }
     return;
}

//////////////////
void KeyFinp(void)
{
     switch(Regim)
     {
       case FINPUT:
       case  RADIO:
		    Mod2Lcd();

		    memset((void *)&LcdBuf[Lcd4f],0,6);

		    LcdBlink  = 0x0008;
		    LcdBuf[    1] = Lcd_F;
		    LcdBuf[Lcd4f] = Lcd__;

		    InpInd = 0;
		    Regim  = FINPUT;
		    break;

       case  WR_MEM:
		    cMEM();
		    LcdBlink = 0x01FF; sleep(2000);
		    LcdBlink = 0x0003;
     }
     return;
}

////////////////////////
void Key_ii( uint8_t c )
{
  uint32_t f;

	LcdBlink <<= 1;

	LcdBuf[InpInd+Lcd4f] &= Lcd_Comma;
	LcdBuf[InpInd+Lcd4f] |= LcdDigit(c);;

	InpBuf[InpInd++] = c +'0';
	InpBuf[InpInd  ] = 0;

     if(InpInd >= InpBufLEN-1)
     {
	f = a2l((uint8_t *)InpBuf);

	SetFwork(f);
     }
     else
	LcdBuf[InpInd+Lcd4f] = Lcd__;
}

////////////////
void Key_0(void) { Key_ii(0); return; }
void Key_1(void) { Key_ii(1); return; }
void Key_2(void) { Key_ii(2); return; }
void Key_3(void) { Key_ii(3); return; }
void Key_4(void) { Key_ii(4); return; }
void Key_5(void) { Key_ii(5); return; }
void Key_6(void) { Key_ii(6); return; }
void Key_7(void) { Key_ii(7); return; }
void Key_8(void) { Key_ii(8); return; }
void Key_9(void) { Key_ii(9); return; }



///////////////////////////////////////
void KeySetHL(uint16_t addr, uint8_t c)
{
     switch(Regim)
     {
       case WR_H_L:
		    wBufEEPROM (addr, (uint8_t *)&FrcvWork, 4);

		    LcdBlink  = 0;
		    LcdBuf[1] = c;

		    sleep(2000);
		    KeyRADIO();
		    break;
       case  RADIO:
		    LcdBlink  = 2;
		    LcdBuf[1] = c;

		    Regim  = WR_H_L;
     }
     return;
}

///////////////////
void KeySetHi(void) { KeySetHL(ADR_EE_Hi,Lcd_H); return; }
void KeySetLo(void) { KeySetHL(ADR_EE_Lo,Lcd_L); return; }

/////////////////////////////////////
uint8_t buff[8]; // for min size code
/////////////////////////////////////

////////////////
void wMEM(void)
{
     memcpy(buff,(uint8_t *)&FrcvWork,4);

     buff[3] = Modulation | 0xFC;

     wBufEEPROM (((uint16_t)IndMEM)<<2, buff, 4);
}

//////////////////
uint8_t rMEM(void)
{
     rBufEEPROM (((uint16_t)IndMEM)<<2, buff, 4);

     if((buff[3] & 0x03) != 0x03)
     {
	Modulation = buff[3] & 0x03;

	memcpy((uint8_t *)&FrcvWork,buff,3);
	SetFrcv();
	Mod2Lcd();
     }
     else
     {
	 memset((void *)&LcdBuf[Lcd4f],Lcd_Minus,6);
	 return 1;
     }
     return 0;
}

////////////////
void cMEM(void)
{
     buff[0] =
     buff[1] =
     buff[2] =
     buff[3] = 0xff;

     wBufEEPROM (((uint16_t)IndMEM)<<2, buff, 4);
}

///////////////////
void KeyWrMem(void)
{
     switch(Regim)
     {
       case WR_MEM:
		    wMEM();
		    LcdBlink = 0;

		    sleep(2000);
		    KeyRADIO();
		    break;
       case  RADIO:
		    LcdBlink = 0x0003;
		    Ind2Lcd();

		    Regim  = WR_MEM;
     }
     return;
}

///////////////////
void KeyRcMem(void)
{
     if(Regim == RADIO)
     {
	Regim = RC_MEM;

	Ind2Lcd();
	rMEM();
     }
     return;
}

///////////////////
void KeyScanF(void)
{
     if(Regim == SCAN_F)
     {
	FlagScanF = (FlagScanF)? 0:1;
     }

     if(Regim == RADIO)
     {
	Regim = SCAN_F;

	LcdBuf[1] = Lcd_Poisk;  // "POISK"
	Mod2Lcd();

	rBufEEPROM (ADR_EE_Hi, (uint8_t *)&FrcvMax, 4);
	rBufEEPROM (ADR_EE_Lo, (uint8_t *)&FrcvMin, 4);

	if(FrcvMax < FrcvMin)
	{
	  rBufEEPROM (ADR_EE_Lo, (uint8_t *)&FrcvMax, 4);
	  rBufEEPROM (ADR_EE_Hi, (uint8_t *)&FrcvMin, 4);
	}
	if(FrcvMax & 0x80000000) SetFmax();
	if(FrcvMin & 0x80000000) SetFmin();

	if( FrcvWork > FrcvMax ||
	    FrcvWork < FrcvMin || FlagBegin)
	{
	    FrcvWork = FrcvMin;
	    FlagBegin= 0;
	    SetFrcv();
	}
	FlagScanF = 1;
     }
     return;
}

//////////////////
void ScanerF(void)
{
     FrcvUpDown(); SetFrcv();
}

///////////////////
void KeyScanM(void)
{
     if(Regim == SCAN_M)
     {
	if(FlagBank)
	{
	   FlagScanM = (FlagScanM)? 0:1;
	}
	else
	{
	   FlagScanM = 1;
	   FlagBank  = 1;
	}
     }

     if(Regim == RADIO)
     {
	Regim = SCAN_M;

	Str2Lcd(sBanc);
	Int2Lcd(ScanMBank,Lcd4i);
	FlagBank = 0;
     }
     return;
}

//////////////////
void ScanerM(void)
{
  static uint8_t ScanMmin[4] = { 1,25,50,75};
  static uint8_t ScanMmax[4] = {24,49,74,99};
  uint8_t IndTmp;

     if( IndMEM < ScanMmin[ScanMBank] ||
	 IndMEM > ScanMmax[ScanMBank] )
	 IndMEM = ScanMmin[ScanMBank];

     IndTmp = IndMEM;

     do
     {
	if(++IndMEM > ScanMmax[ScanMBank])
	{
	     IndMEM = ScanMmin[ScanMBank];
	}

	if(!rMEM())
	{
	   SmeterPrn();
	   SmeterPrn();
	   SmeterPrn();
	   SmeterPrn();
	   SmeterPrn();
	   SmeterPrn();
	   break;
	}

     } while(IndTmp != IndMEM);

     Ind2Lcd();
     return;
}


  /////////////////////
  /////////////////////
  volatile uint8_t SQL;

////////////////
void Sql2Lcd(void)
{
     Int2Lcd(SQL,Lcd4i);  TimeClr = AUTO_CLR;
}

////////////////
void SQLup(void)
{
     if(SQL < 255)
	SQL++;

     Sql2Lcd();
}

//////////////////
void SQLdown(void)
{
     if(SQL)
	SQL--;

     Sql2Lcd();
}

///////////////////
void KeySQL(void)
{
     if(Regim == RADIO)
     {
	Regim = SET_SQL;

	Str2Lcd(sSql);
	Sql2Lcd();
     }
     return;
}

///////////////////
void SqlRADIO(void)
{
   if( Regim != SET_SQL && Regim != RADIO )
   {
       KeyRADIO();
   }
   return;
}

///////////////////
void VolRADIO(void)
{
   if( Regim != SET_VOL && Regim != RADIO )
   {
       KeyRADIO();
   }
   return;
}

//////////////////////////////////////////////////////////////////
void PDUsqlUp  (void) { SqlRADIO(); KeySQL(); SQLup();   return; }
void PDUsqlDown(void) { SqlRADIO(); KeySQL(); SQLdown(); return; }

//////////////////
void SetMute(void)
{
     if(FlagSQL == 0)
	return;

      if(Smeter > SQL || SQL == 0)
      {
      // Sound ON

	 FlagScanF = 0;
	 FlagScanM = 0;
	 MuteCnt = 0;

	 if(FlagMute)
	 {
	    FlagMute = 0;
	    Lc72131Set();
	 }
      }
      else
      {
      // Sound Off

	 if(MuteCnt >= MUTE)
	 {
	    if(!FlagMute)
	    {
		FlagMute = 1;
		Lc72131Set();
	    }
	 }
      }
      return;
}

void RsControl(void)
{
  uint8_t inp;
  uint8_t   i;
  uint32_t  f;

  if(Regim == RADIO || Regim == SCAN_F)
  {
     if(RxInd != TxInd)
     {
	if((inp = RxInpBuf[TxInd++]) != '\n' && EchoYes)
	{
	   uart_putchar(inp);
	}
	//uart_putchar(inp = RxInpBuf[TxInd++]);

	if(RxInd >= RxInpBufLEN)
	{
	   RcInd = 0;
	   RxInd = 0;
	   TxInd = 0;
	   uart_putstr(ERROR);

	   return;
	}
	if(isdigit(inp))
	{
	   RcInpBuf[RcInd++] = inp;
	}

	if(inp == '\r')
	{
	   RcInpBuf[RcInd] = '\0';
	   RcInd = 0;

	   if(Regim == SCAN_F)
	   {
	      RxInpBuf[0] = 0;
	      KeyRADIO();
	   }
	   f = a2l((uint8_t *)RcInpBuf);

	   i = (uint8_t)(f & 0x000000FF);

	   // use uint8_t inp like tmp

	   inp = (RxInpBuf[0] == '\n')? 1:0;

	   switch(RxInpBuf[inp])
	   {
	    case '+': KeyPlus (); break;
	    case '-': KeyMinus(); break;

	    case 'M':
		      if((Modulation = i & 0x03) == 3)
			  Modulation = 0;
		      Mod2Lcd();
		      Lc72131Set();
		      break;

	    case 'G': TunerGain = i & 0x01;
		      KSH14xPLL();
		      break;

	    case 'S': FindStep = i &0x07;
		      FrcvStep = Step[FindStep];
		      break;

	    case 'U': Uapy    = i;
		      UapyFlg = 2; UapyOut();
		      break;

	    case 'E': EchoYes = i;
		      break;

	    case 'L':
		      if(i == 2)
		      {
			 FlagSQL = 1;
		      }
		      else
		      {
			 FlagMute = i & 0x01;
			 FlagSQL  = 0;
			 Lc72131Set();
		      }
		      break;

#ifdef RC5_CONTROL
	    case 'R': RC5wrk = i & 0x01;
		      break;

	    case 'W': RC5write(i);
		      break;
#endif
	    case '.': fWorkPrn();
		      break;

	    case ',': SmeterPrn();
		      SmeterPrn();
		      SmeterPrn();
		      SmeterPrn();
		      SmeterPrn();
		      break;

	    case 'D': if(i)
		      {
			LcdWork = 1;
		      }
		      else
		      {
			/*
			LcdWork = 0;
			cli();
			LcdOut2(0);
			LcdOut2(0);
			sei();
			*/
		      }
		      break;

	    case 'H': wBufEEPROM (ADR_EE_Hi, (uint8_t *)&f, 4);
		      break;

	    case 'B': wBufEEPROM (ADR_EE_Lo, (uint8_t *)&f, 4);
		      break;


	    case 'F': SetFwork(f);
		      break;

	    case 'P': KeyScanF();
		      break;

	    case 'N': FlagBegin++;
		      KeyScanF();
		      break;

	    case 'V': wBufEEPROM (ADR_EE_vMax, (uint8_t *)&i, 1);
		      wBufEEPROM (ADR_EE_vWrk, (uint8_t *)&i, 1); break;
	    case 'v': wBufEEPROM (ADR_EE_vMin, (uint8_t *)&i, 1); break;

	   }
	   RxInd = 0;
	   TxInd = 0;
	   uart_putstr(PROMPT);
	}
     }
  }
  return;
}

//////////////////////
void UapyControl(void)
{
   if(UapyFlg == 1)
   {
     SmeterUapy += (uint16_t)Smeter;

     if(++UapyCnt >= UapyNN)
     {
	SmeterUapy = SmeterUapy / UapyNN;

	if(SmeterUapy > UapyHi)
	{
	  if(Uapy > 100)
	  {
	     Uapy--; UapyOut();
	  }
	}
	if(SmeterUapy < UapyLo)
	{
	  if(Uapy < 255)
	  {
	     Uapy++; UapyOut();
	  }

	}
	SmeterUapy = 0;
	UapyCnt    = 0;
     }
   }
   return;
}

///////////////////
void Vol2Lcd(void)
{
     Int2Lcd((uint16_t)(Vol_Max-Pwm1),Lcd4i);
     TimeClr = AUTO_CLR;
}

///////////////////
void VolumeUp(void)
{
   if(Pwm1 > Vol_Min)
   {
     OCR1A = Pwm1--;
   }
   Vol2Lcd();
   return;
}

/////////////////////
void VolumeDown(void)
{
   if(Pwm1 < Vol_Max)
   {
     OCR1A = Pwm1++;
   }
   Vol2Lcd();
   return;
}

///////////////////
void KeyVOL(void)
{
     if(Regim == RADIO)
     {
	Regim = SET_VOL;

	Str2Lcd(sVol);
	Vol2Lcd();
     }
     return;
}

/////////////////////////////////////////////////////////////////////
void PDUvolUp  (void) { VolRADIO(); KeyVOL(); VolumeUp();   return; }
void PDUvolDown(void) { VolRADIO(); KeyVOL(); VolumeDown(); return; }

//////////////////////////////////////
//////////////////////////////////////

	  uint8_t l2aBuff[8];

//////////////////////////////////////
uint8_t * ul2a(uint32_t c,uint8_t len)
{
  uint8_t  i;
  uint8_t *p;

   p = &l2aBuff[len];
  *p = 0;

  for(i = len; i > 0; i--)
  {
      p--;
     *p  = c%10 +'0';
      c /= 10;
  }
  return p;
}

////////////////////////
uint32_t a2l(uint8_t *p)
{
  uint32_t l = 0;

  while(*p)
  {
     l *= 10;
     l += *p-'0';
     p++;
  }
  return l;
}

//////////////////////////////////////
//                                  //
//  MT10T7-x , MT10T8-x , MT10T9-x  //
//                                  //
//////////////////////////////////////

#define MT10Tx_A0       EP7
#define MT10Tx_WR       EP6
#define MT10Tx_D3       EP5
#define MT10Tx_D2       EP4
#define MT10Tx_D1       EP3
#define MT10Tx_D0       EP2

#define MT10Tx_Lamp     EP1

#define MT10Tx_PORT     ExtPort

#define HiBits(w) ((w >> 4) & 0x0f)
#define LoBits(w) ((w >> 0) & 0x0f)

///////////////////////
void MT10TxLampOn(void)
{
     SetExtBit(MT10Tx_Lamp);
}

///////////////////////
void MT10TxLampOff(void)
{
     ClrExtBit(MT10Tx_Lamp);
}

/////////////////////
void MT10TxWrite(void)
{
     SetExtBit(MT10Tx_WR);
     ClrExtBit(MT10Tx_WR);
}

////////////////////////////
void MT10TxOut(uint8_t data)
{
  uint8_t i=0;

     MT10Tx_PORT &= ~(_BV(MT10Tx_D0)|
		      _BV(MT10Tx_D1)|
		      _BV(MT10Tx_D2)|
		      _BV(MT10Tx_D3));

     if(data & 0x01) i |= _BV(MT10Tx_D0);
     if(data & 0x02) i |= _BV(MT10Tx_D1);
     if(data & 0x04) i |= _BV(MT10Tx_D2);
     if(data & 0x08) i |= _BV(MT10Tx_D3);
     MT10Tx_PORT |= i;

     ExtPortSet ();
     MT10TxWrite();
}

////////////////////////////
void MT10TxSet(uint8_t data)
{
     ClrBit(MT10Tx_PORT,MT10Tx_A0); MT10TxOut(data);
     SetBit(MT10Tx_PORT,MT10Tx_A0);
}
/////////////////////
void MT10TxInit(void)
{
  uint8_t i;

     MT10TxSet(0x0f);
     MT10TxOut(0x01);
     MT10TxSet(0x00);

     for(i = 0; i < 20; i++)
     {
	 MT10TxOut(0);
     }
     MT10TxLampOn();
}

////////////////////
void MT10TxLcd(void)
{
  uint8_t i,c;

  if(LcdYes)
  {
     LcdYes = 0;

     MT10TxSet(0x00);

     for(i = 0; i < 9; i++)
     {
       c = LcdBuf[i];

       if((LcdBlink >> i) & 0x0001)
       {
	   if(LcdFlg) c = 0;
       }
       MT10TxOut(LoBits(c));
       MT10TxOut(HiBits(c));
     }
  }
}
